from sqlalchemy import Integer, Float, Boolean, Column
from graphql import (
    GraphQLString,
    GraphQLInt,
    GraphQLFloat,
    GraphQLBoolean,
    GraphQLInputField,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
)


def get_graphql_type_from_column(column: Column) -> GraphQLScalarType:
    if isinstance(column.type, Integer):
        return GraphQLInt

    if isinstance(column.type, Float):
        return GraphQLFloat

    if isinstance(column.type, Boolean):
        return GraphQLBoolean

    return GraphQLString


def get_base_comparison_fields(scalar: GraphQLScalarType):
    return {
        "_eq": GraphQLInputField(scalar),
        "_neq": GraphQLInputField(scalar),
        "_in": GraphQLInputField(GraphQLList(GraphQLNonNull(scalar))),
        "_nin": GraphQLInputField(GraphQLList(GraphQLNonNull(scalar))),
        "_lt": GraphQLInputField(scalar),
        "_gt": GraphQLInputField(scalar),
        "_gte": GraphQLInputField(scalar),
        "_lte": GraphQLInputField(scalar),
        "_is_null": GraphQLInputField(GraphQLBoolean),
    }


def get_string_comparison_fields():
    return {"_like": GraphQLInputField(GraphQLString), "_nlike": GraphQLInputField(GraphQLString)}



from typing import Dict

from sqlalchemy.ext.declarative.api import DeclarativeMeta
from graphql import (
    GraphQLObjectType,
    GraphQLField,
    GraphQLSchema,
    GraphQLList,
    GraphQLInputObjectType,
)

from .resolvers import make_resolver, make_pk_resolver
from .inputs import make_args, make_pk_args
from .names import get_model_pk_field_name
from .objects import build_table_type


def build_schema(base: DeclarativeMeta) -> GraphQLSchema:
    fields: Dict[str, GraphQLField] = {}
    objects: Dict[str, GraphQLObjectType] = {}
    inputs: Dict[str, GraphQLInputObjectType] = {}

    for model in base.__subclasses__():
        object_type = build_table_type(model, objects)

        objects[object_type.name] = object_type
        fields[object_type.name] = GraphQLField(
            GraphQLList(object_type), args=make_args(model, inputs=inputs), resolve=make_resolver(model)
        )

        pk_field_name = get_model_pk_field_name(model)
        fields[pk_field_name] = GraphQLField(object_type, args=make_pk_args(model), resolve=make_pk_resolver(model))

    return GraphQLSchema(GraphQLObjectType("Query", fields))



from typing import Optional, Dict

from graphql import (
    GraphQLArgument,
    GraphQLNonNull,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLList,
    GraphQLEnumType,
    GraphQLInputField,
)
from sqlalchemy.ext.declarative import DeclarativeMeta

from .scalars import get_graphql_type_from_column
from .objects import get_comparison_object_type
from .names import get_model_order_by_input_name, get_model_where_input_name


PAGINATION_ARGS = {"limit": GraphQLInt, "offset": GraphQLInt}
ORDER_BY_ENUM = GraphQLEnumType("order_by", {"desc": "desc", "asc": "asc"})


def make_where_type(model: DeclarativeMeta, inputs: Dict[str, GraphQLInputObjectType]) -> GraphQLInputObjectType:
    type_name = get_model_where_input_name(model)

    def get_fields():
        fields = {
            "_and": GraphQLList(inputs[type_name]),
            "_or": GraphQLList(inputs[type_name]),
            "_not": inputs[type_name],
        }

        for column in model.__table__.columns:  # type: ignore
            fields[column.name] = GraphQLInputField(get_comparison_object_type(column, inputs))

        for name, relationship in model.__mapper__.relationships.items():
            fields[name] = inputs[get_model_where_input_name(relationship.mapper.entity)]

        return fields

    input_type = GraphQLInputObjectType(type_name, get_fields)
    inputs[type_name] = input_type
    return input_type


def make_order_type(model: DeclarativeMeta, inputs: Dict[str, GraphQLInputObjectType]) -> GraphQLInputObjectType:
    type_name = get_model_order_by_input_name(model)

    def get_fields():
        fields = {}

        for column in model.__table__.columns:  # type: ignore
            fields[column.name] = GraphQLInputField(ORDER_BY_ENUM)

        for name, relationship in model.__mapper__.relationships.items():
            fields[name] = inputs[get_model_order_by_input_name(relationship.mapper.entity)]

        return fields

    input_type = GraphQLInputObjectType(type_name, get_fields)
    inputs[type_name] = input_type
    return input_type


def make_args(model: DeclarativeMeta, inputs: Dict[str, GraphQLInputObjectType]) -> Dict[str, GraphQLArgument]:
    args = {}
    for name, field in PAGINATION_ARGS.items():
        args[name] = GraphQLArgument(field)

    order_type = make_order_type(model, inputs)
    args["order"] = GraphQLArgument(GraphQLList(GraphQLNonNull(order_type)))

    where_type = make_where_type(model, inputs)
    args["where"] = GraphQLArgument(where_type)

    return args


def make_pk_args(model: DeclarativeMeta) -> Optional[Dict[str, GraphQLArgument]]:
    primary_key = model.__table__.primary_key  # type: ignore

    if not primary_key:
        return None

    args = {}
    for column in primary_key.columns:
        graphql_type = get_graphql_type_from_column(column)
        args[column.name] = GraphQLArgument(GraphQLNonNull(graphql_type))

    return args



from sqlalchemy.ext.declarative import DeclarativeMeta
from graphql import GraphQLScalarType


def get_table_name(model: DeclarativeMeta) -> str:
    return model.__tablename__  # type: ignore


def get_model_pk_field_name(model: DeclarativeMeta) -> str:
    return f"{get_table_name(model)}_by_pk"


def get_model_order_by_input_name(model: DeclarativeMeta) -> str:
    return f"{get_table_name(model)}_order_by"


def get_model_where_input_name(model: DeclarativeMeta) -> str:
    return f"{get_table_name(model)}_bool_exp"


def get_scalar_comparison_name(scalar: GraphQLScalarType) -> str:
    return f"{scalar.name.lower()}_comparison_exp"



from functools import partial
from itertools import starmap
from typing import Any, Callable, Dict, List, Union

from sqlalchemy import Column, or_, and_, not_, true
from sqlalchemy.sql import ClauseElement
from sqlalchemy.orm import Query
from sqlalchemy.ext.declarative import DeclarativeMeta


def make_field_resolver(field: str) -> Callable:
    def resolver(root: DeclarativeMeta, _info: Any):
        return getattr(root, field)

    return resolver


def get_bool_operation(model_property: Column, operator: str, value: Any) -> Union[bool, ClauseElement]:
    if operator == "_eq":
        return model_property == value

    if operator == "_in":
        return model_property.in_(value)

    if operator == "_is_null":
        return model_property.is_(None)

    if operator == "_like":
        return model_property.like(value)

    if operator == "_neq":
        return model_property != value

    if operator == "_nin":
        return model_property.notin_(value)

    if operator == "_nlike":
        return model_property.notlike(value)

    if operator == "_lt":
        return model_property < value

    if operator == "_gt":
        return model_property > value

    if operator == "_lte":
        return model_property <= value

    if operator == "_gte":
        return model_property >= value

    raise Exception("Invalid operator")


def get_filter_operation(model: DeclarativeMeta, where: Dict[str, Any]) -> ClauseElement:
    partial_filter = partial(get_filter_operation, model)

    for name, exprs in where.items():
        if name == "_or":
            return or_(*map(partial_filter, exprs))

        if name == "_not":
            return not_(partial_filter(exprs))

        if name == "_and":
            return and_(*map(partial_filter, exprs))

        model_property = getattr(model, name)
        partial_bool = partial(get_bool_operation, model_property)
        return and_(*(starmap(partial_bool, exprs.items())))

    return true()


def filter_query(model: DeclarativeMeta, query: Query, where: Dict[str, Any] = None) -> Query:
    if not where:
        return query

    query_filter = getattr(query, "filter")
    for name, exprs in where.items():
        query = query_filter(get_filter_operation(model, {name: exprs}))

    return query


def order_query(model: DeclarativeMeta, query: Query, order: List[Dict[str, Any]] = None) -> Query:
    if not order:
        return query

    for expr in order:
        for name, direction in expr.items():
            model_property = getattr(model, name)
            model_order = getattr(model_property, direction)
            query_order = getattr(query, "order_by")
            query = query_order(model_order())

    return query


def make_resolver(model: DeclarativeMeta) -> Callable:
    def resolver(
        _root: DeclarativeMeta,
        info: Any,
        where: Dict[str, Any] = None,
        order: List[Dict[str, Any]] = None,
        limit: int = None,
        offset: int = None,
    ):
        session = info.context["session"]
        query = session.query(model)

        query = filter_query(model, query, where)
        query = order_query(model, query, order)

        if limit:
            query = getattr(query, "limit")(limit)

        if offset:
            query = getattr(query, "offset")(offset)

        return query.all()

    return resolver


def make_pk_resolver(model: DeclarativeMeta) -> Callable:
    def resolver(_root: DeclarativeMeta, info: Any, **kwargs: Dict[str, Any]):
        session = info.context["session"]
        return session.query(model).get(kwargs)

    return resolver



from typing import Dict

from sqlalchemy import Column
from sqlalchemy.orm import interfaces
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from graphql import (
    GraphQLObjectType,
    GraphQLField,
    GraphQLList,
    GraphQLInputObjectType,
    GraphQLString,
)

from .resolvers import make_field_resolver
from .scalars import get_graphql_type_from_column, get_base_comparison_fields, get_string_comparison_fields
from .names import get_table_name, get_scalar_comparison_name


def build_table_type(model: DeclarativeMeta, objects: Dict[str, GraphQLObjectType]) -> GraphQLObjectType:
    table = model.__table__  # type: ignore

    def get_fields():
        fields = {}

        for column in table.columns:
            graphql_type = get_graphql_type_from_column(column)
            fields[column.name] = GraphQLField(graphql_type, resolve=make_field_resolver(column.name))

        for name, relationship in model.__mapper__.relationships.items():
            object_type = objects[get_table_name(relationship.mapper.entity)]
            if relationship.direction in (interfaces.ONETOMANY, interfaces.MANYTOMANY):
                object_type = GraphQLList(object_type)

            fields[name] = fields[name] = GraphQLField(object_type, resolve=make_field_resolver(name),)

        return fields

    return GraphQLObjectType(get_table_name(model), get_fields)


def get_comparison_object_type(column: Column, inputs: Dict[str, GraphQLInputObjectType]) -> GraphQLInputObjectType:
    scalar = get_graphql_type_from_column(column)
    type_name = get_scalar_comparison_name(scalar)

    if type_name in inputs:
        return inputs[type_name]

    fields = get_base_comparison_fields(scalar)

    if scalar == GraphQLString:
        fields.update(get_string_comparison_fields())

    object_type = GraphQLInputObjectType(type_name, fields)
    inputs[type_name] = object_type

    return object_type



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['graphql_sqlalchemy']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.2,<2', 'graphql-core>=3.0.0,<4']

setup_kwargs = {
    'name': 'graphql-sqlalchemy',
    'version': '0.4.0',
    'description': 'Generate GraphQL Schemas from your SQLAlchemy models',
    'long_description': '# graphql-sqlalchemy\n\n[![PyPI version](https://badge.fury.io/py/graphql-sqlalchemy.svg)](https://badge.fury.io/py/graphql-sqlalchemy)\n[![Build Status](https://travis-ci.com/gzzo/graphql-sqlalchemy.svg?branch=master)](https://travis-ci.com/gzzo/graphql-sqlalchemy)\n[![codecov](https://codecov.io/gh/gzzo/graphql-sqlalchemy/branch/master/graph/badge.svg)](https://codecov.io/gh/gzzo/graphql-sqlalchemy)\n[![Code Style](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\nGenerate GraphQL Schemas from your SQLAlchemy models\n\n# Install\n```\npip install graphql-sqlalchemy\n```\n\n# Usage\n\n```python\nfrom ariadne import QueryType\nfrom ariadne.asgi import GraphQL\nfrom fastapi import FastAPI\nfrom graphql_sqlalchemy import build_schema\n\nfrom .session import Session\nfrom .base import Base\n\n\napp = FastAPI()\nquery = QueryType()\nsession = Session()\n\nschema = build_schema(Base)\nquery.bind_to_schema(schema)\n\napp.mount("/graphql", GraphQL(schema, context_value=dict(session=session)))\n```\n',
    'author': 'Guido Rainuzzo',
    'author_email': 'hi@guido.nyc',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gzzo/graphql-sqlalchemy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.0,<4',
}


setup(**setup_kwargs)
